#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#
# Author: Komal Thareja (kthare10@renci.org)
import json
import logging
import os
from datetime import datetime, timedelta, timezone
from typing import Tuple, Union, List, Any, Dict

import paramiko
from fabric_cf.orchestrator.swagger_client import Sliver, Slice
from fabric_cf.orchestrator.swagger_client.models import PoaData
from fabric_cm.credmgr.credmgr_proxy import TokenType

from fabrictestbed.external_api.core_api import CoreApi
from fabrictestbed.slice_editor import ExperimentTopology, AdvertisedTopology, Node, GraphFormat
from fabrictestbed.slice_manager import CredmgrProxy, OrchestratorProxy, CmStatus, Status, SliceState
from fabrictestbed.util.constants import Constants
from fabrictestbed.util.utils import Utils


class SliceManagerException(Exception):
    """ Slice Manager Exception """


class SliceManager:
    """
    Implements User facing Control Framework API interface
    """
    def __init__(self, *, cm_host: str = None, oc_host: str = None, core_api_host: str = None,
                 token_location: str = None,
                 project_id: str = None, scope: str = "all", initialize: bool = True,
                 project_name: str = None, auto_refresh: bool = True):
        self.auto_refresh = auto_refresh
        self.logger = logging.getLogger()
        if cm_host is None:
            cm_host = os.environ.get(Constants.FABRIC_CREDMGR_HOST)
        if oc_host is None:
            oc_host = os.environ.get(Constants.FABRIC_ORCHESTRATOR_HOST)
        if core_api_host is None:
            core_api_host = os.environ.get(Constants.FABRIC_CORE_API_HOST)
        self.cm_proxy = CredmgrProxy(credmgr_host=cm_host)
        self.oc_proxy = OrchestratorProxy(orchestrator_host=oc_host)
        self.core_api_host = core_api_host
        self.token_location = token_location
        self.tokens = {}
        self.project_id = project_id
        if self.project_id is None:
            self.project_id = os.environ.get(Constants.FABRIC_PROJECT_ID)
        self.project_name = project_name
        if self.project_name is None:
            self.project_name = os.environ.get(Constants.FABRIC_PROJECT_NAME)
        self.scope = scope
        if self.token_location is None:
            self.token_location = os.environ.get(Constants.FABRIC_TOKEN_LOCATION)
        self.initialized = False

        if cm_host is None or oc_host is None or self.core_api_host is None or self.token_location is None:
            raise SliceManagerException(f"Invalid initialization parameters: cm_host: {cm_host}, "
                                        f"oc_host: {oc_host} core_api_host: {core_api_host} "
                                        f"token_location: {self.token_location}")

        # Try to load the project_id or project_name from the Token
        if project_id is None and project_name is None:
            self.__determine_project(cm_host=cm_host)

        # Validate the required parameters are set
        if self.project_id is None and self.project_name is None:
            raise SliceManagerException(f"Invalid initialization parameters: project_id={self.project_id}, "
                                        f"project_name={self.project_name}")

        if initialize:
            self.initialize()

    def __determine_project(self, cm_host: str):
        self.__load_tokens(refresh=False)
        if self.get_id_token() is not None:
            logging.info("Project Id/Name not specified, trying to determine it from the token")
            decoded_token = Utils.decode_token(cm_host=cm_host, token=self.get_id_token())
            if decoded_token.get("projects") and len(decoded_token.get("projects")):
                self.project_id = decoded_token.get("projects")[0].get("uuid")
                self.project_name = decoded_token.get("projects")[0].get("name")

    def initialize(self):
        """
        Initialize the Slice Manager object
        - Load the tokens
        - Refresh if needed
        """
        if not self.initialized:
            self.__load_tokens()
            self.initialized = True

    def __check_initialized(self):
        """
        Check if Slice Manager has been initialized
        @raises Exception if slice manager has been initialized
        """
        if not self.initialized:
            raise SliceManagerException("Slice Manager has not been initialized!")

    def __should_renew(self) -> bool:
        """
        Check if tokens should be renewed
        Returns true if tokens are at least 30 minutes old
        @return true if tokens should be renewed; false otherwise
        """
        self.__check_initialized()

        id_token = self.get_id_token()
        created_at = self.tokens.get(CredmgrProxy.CREATED_AT, None)

        created_at_time = datetime.strptime(created_at, CredmgrProxy.TIME_FORMAT)
        now = datetime.now(timezone.utc)

        if id_token is None or now - created_at_time >= timedelta(minutes=180):
            return True

        return False

    def __load_tokens(self, refresh: bool = True):
        """
        Load Fabric Tokens from the tokens.json if it exists
        Otherwise, this is the first attempt, create the tokens and save them
        @note this function is invoked when reloading the tokens to ensure tokens
        from the token file are read instead of the local variables
        """
        # Load the tokens from the JSON
        if os.path.exists(self.token_location):
            with open(self.token_location, 'r') as stream:
                self.tokens = json.loads(stream.read())
            refresh_token = self.get_refresh_token()
        else:
            # First time login, use environment variable to load the tokens
            refresh_token = os.environ.get(Constants.CILOGON_REFRESH_TOKEN)
        # Renew the tokens to ensure any project_id changes are taken into account
        if refresh and self.auto_refresh:
            self.refresh_tokens(refresh_token=refresh_token)

    def get_refresh_token(self) -> str:
        """
        Get Refresh Token
        @return refresh token
        """
        return self.tokens.get(CredmgrProxy.REFRESH_TOKEN, None)

    def get_id_token(self) -> str:
        """
        Get Id token
        @return id token
        """
        return self.tokens.get(CredmgrProxy.ID_TOKEN, None)

    def set_token_location(self, *, token_location: str):
        """
        Set token location: path of the file where tokens should be saved
        @param token_location file name along with complete path where tokens should be stored
        """
        self.token_location = token_location

    def create_token(self, scope: str = "all", project_id: str = None, project_name: str = None, file_name: str = None,
                     life_time_in_hours: int = 4, comment: str = "Created via API",
                     browser_name: str = "chrome") -> Tuple[Status, Union[dict, SliceManagerException]]:
        """
        Create token
        @param project_id: Project Id
        @param project_name: Project Name
        @param scope: scope
        @param file_name: File name
        @param life_time_in_hours: Token lifetime in hours
        @param comment: comment associated with the token
        @param browser_name: Browser name; allowed values: chrome, firefox, safari, edge
        @returns Tuple of Status, token json or Exception
        @raises Exception in case of failure
        """
        try:
            return self.cm_proxy.create(scope=scope, project_id=project_id, project_name=project_name,
                                        file_name=file_name, life_time_in_hours=life_time_in_hours, comment=comment,
                                        browser_name=browser_name)
        except Exception as e:
            error_message = Utils.extract_error_message(exception=e)
            return Status.FAILURE, SliceManagerException(error_message)

    def refresh_tokens(self, *, refresh_token: str) -> Tuple[str, str]:
        """
        Refresh tokens
        User is expected to invoke refresh token API before invoking any other APIs to ensure the token is not expired.
        User is also expected to update the returned refresh token in the JupyterHub environment.
        @returns tuple of id token and refresh token
        @note this exposes an API for the user to refresh tokens explicitly only. CredMgrProxy::refresh already
        updates the refresh tokens to the token file atomically.
        """
        try:
            status, tokens = self.cm_proxy.refresh(project_id=self.project_id, scope=self.scope,
                                                   refresh_token=refresh_token, file_name=self.token_location,
                                                   project_name=self.project_name)
            if status == CmStatus.OK:
                self.tokens = tokens
                return tokens.get(CredmgrProxy.ID_TOKEN, None), tokens.get(CredmgrProxy.REFRESH_TOKEN, None)
            else:
                error_message = Utils.extract_error_message(exception=tokens)
                raise SliceManagerException(error_message)
        except Exception as e:
            error_message = Utils.extract_error_message(exception=e)
            raise SliceManagerException(error_message)

    def revoke_token(self, *, refresh_token: str = None, id_token: str = None, token_hash: str = None,
                     token_type: TokenType = TokenType.Refresh) -> Tuple[Status, Any]:
        """
        Revoke a refresh token
        @param refresh_token Refresh Token to be revoked
        @param id_token Identity Token
        @param token_hash Token Hash
        @param token_type type of the token being revoked
        @return Tuple of the status and revoked refresh token
        """
        if refresh_token is None:
            refresh_token = self.get_refresh_token()
        if id_token is None:
            id_token = self.get_id_token()
        if token_hash is None:
            token_hash = Utils.generate_sha256(token=id_token)

        try:
            return self.cm_proxy.revoke(refresh_token=refresh_token, identity_token=id_token, token_hash=token_hash,
                                        token_type=token_type)
        except Exception as e:
            error_message = Utils.extract_error_message(exception=e)
            return Status.FAILURE, SliceManagerException(error_message)

    def token_revoke_list(self, *, project_id: str) -> Tuple[Status, Union[SliceManagerException, List[str]]]:
        """
        Get Token Revoke list for a project
        @param project_id project_id
        @return token revoke list
        """
        try:
            return self.cm_proxy.token_revoke_list(project_id=project_id)
        except Exception as e:
            error_message = Utils.extract_error_message(exception=e)
            return Status.FAILURE, SliceManagerException(error_message)

    def clear_token_cache(self, *, file_name: str = None):
        """
        Clear the cached token
        Should be invoked when the user changes projects
        @return:
        """
        cache_file_name = file_name
        if cache_file_name is None:
            cache_file_name = self.token_location
        status, exception = self.cm_proxy.clear_token_cache(file_name=cache_file_name)
        if status == CmStatus.OK:
            return Status.OK, None
        return Status.FAILURE, f"Failed to clear token cache: {Utils.extract_error_message(exception=exception)}"

    def create(self, *, slice_name: str, ssh_key: Union[str, List[str]], topology: ExperimentTopology = None,
               slice_graph: str = None, lease_start_time: str = None,
               lease_end_time: str = None) -> Tuple[Status, Union[SliceManagerException, List[Sliver]]]:
        """
        Create a slice
        @param slice_name slice name
        @param ssh_key SSH Key(s)
        @param topology Experiment topology
        @param slice_graph Slice Graph string
        @param lease_start_time Lease Start Time
        @param lease_end_time Lease End Time
        @return Tuple containing Status and Exception/Json containing slivers created
        """
        if slice_name is None or not isinstance(slice_name, str) or ssh_key is None:
            return Status.INVALID_ARGUMENTS, SliceManagerException("Invalid arguments - slice_name or ssh key")

        if topology is not None and not isinstance(topology, ExperimentTopology):
            return Status.INVALID_ARGUMENTS, SliceManagerException("Invalid arguments - topology")

        if slice_graph is not None and not isinstance(slice_graph, str):
            return Status.INVALID_ARGUMENTS, SliceManagerException("Invalid arguments - slice_graph")

        if lease_end_time is not None and not isinstance(lease_end_time, str):
            return Status.INVALID_ARGUMENTS, SliceManagerException("Invalid arguments - lease_end_time")

        try:
            if self.__should_renew():
                self.__load_tokens()
            return self.oc_proxy.create(token=self.get_id_token(), slice_name=slice_name, ssh_key=ssh_key,
                                        topology=topology, slice_graph=slice_graph, lease_end_time=lease_end_time,
                                        lease_start_time=lease_start_time)
        except Exception as e:
            error_message = Utils.extract_error_message(exception=e)
            return Status.FAILURE, SliceManagerException(error_message)

    def modify(self, *, slice_id: str, topology: ExperimentTopology = None,
               slice_graph: str = None) -> Tuple[Status, Union[SliceManagerException, List[Sliver]]]:
        """
        Modify an existing slice
        @param slice_id slice id
        @param topology Experiment topology
        @param slice_graph Slice Graph string
        @return Tuple containing Status and Exception/Json containing slivers created
        """
        if slice_id is None or not isinstance(slice_id, str):
            return Status.INVALID_ARGUMENTS, SliceManagerException("Invalid arguments - slice_id")

        if topology is not None and not isinstance(topology, ExperimentTopology):
            return Status.INVALID_ARGUMENTS, SliceManagerException("Invalid arguments - topology")

        if slice_graph is not None and not isinstance(slice_graph, str):
            return Status.INVALID_ARGUMENTS, SliceManagerException("Invalid argument - slice_graph")

        try:
            if self.__should_renew():
                self.__load_tokens()
            return self.oc_proxy.modify(token=self.get_id_token(), slice_id=slice_id, topology=topology,
                                        slice_graph=slice_graph)
        except Exception as e:
            error_message = Utils.extract_error_message(exception=e)
            return Status.FAILURE, SliceManagerException(error_message)

    def modify_accept(self, *, slice_id: str) -> Tuple[Status, Union[SliceManagerException, ExperimentTopology]]:
        """
        Modify an existing slice
        @param slice_id slice id
        @return Tuple containing Status and Exception/Json containing slivers created
        """
        if slice_id is None or not isinstance(slice_id, str):
            return Status.INVALID_ARGUMENTS, SliceManagerException("Invalid arguments - slice_id")

        try:
            if self.__should_renew():
                self.__load_tokens()
            return self.oc_proxy.modify_accept(token=self.get_id_token(), slice_id=slice_id)
        except Exception as e:
            error_message = Utils.extract_error_message(exception=e)
            return Status.FAILURE, SliceManagerException(error_message)

    def delete(self, *, slice_object: Slice = None) -> Tuple[Status, Union[SliceManagerException, None]]:
        """
        Delete slice(s)
        @param slice_object slice to be deleted
        @return Tuple containing Status and Exception/Json containing deletion status
        """
        try:
            if self.__should_renew():
                self.__load_tokens()
            slice_id = slice_object.slice_id if slice_object is not None else None
            return self.oc_proxy.delete(token=self.get_id_token(), slice_id=slice_id)
        except Exception as e:
            error_message = Utils.extract_error_message(exception=e)
            return Status.FAILURE, SliceManagerException(error_message)

    def slices(self, includes: List[SliceState] = None, excludes: List[SliceState] = None, name: str = None,
               limit: int = 20, offset: int = 0, slice_id: str = None,
               as_self: bool = True) -> Tuple[Status, Union[SliceManagerException, List[Slice]]]:
        """
        Get slices
        @param includes list of the slice state used to include the slices in the output
        @param excludes list of the slice state used to exclude the slices from the output
        @param name name of the slice
        @param limit maximum number of slices to return
        @param offset offset of the first slice to return
        @param slice_id slice id
        @param as_self
        @return Tuple containing Status and Exception/Json containing slices
        """
        try:
            if self.__should_renew():
                self.__load_tokens()
            return self.oc_proxy.slices(token=self.get_id_token(), includes=includes, excludes=excludes,
                                        name=name, limit=limit, offset=offset, slice_id=slice_id, as_self=as_self)
        except Exception as e:
            error_message = Utils.extract_error_message(exception=e)
            return Status.FAILURE, SliceManagerException(error_message)

    def get_slice_topology(self, *, slice_object: Slice, graph_format: GraphFormat = GraphFormat.GRAPHML,
                           as_self: bool = True) -> Tuple[Status, Union[SliceManagerException, ExperimentTopology]]:
        """
        Get slice topology
        @param slice_object Slice for which to retrieve the topology
        @param graph_format
        @param as_self
        @return Tuple containing Status and Exception/Json containing slice
        """
        if slice_object is None or not isinstance(slice_object, Slice):
            return Status.INVALID_ARGUMENTS, SliceManagerException("Invalid arguments - slice_object")
        try:
            if self.__should_renew():
                self.__load_tokens()
            return self.oc_proxy.get_slice(token=self.get_id_token(), slice_id=slice_object.slice_id,
                                           graph_format=graph_format, as_self=as_self)
        except Exception as e:
            error_message = Utils.extract_error_message(exception=e)
            return Status.FAILURE, SliceManagerException(error_message)

    def slivers(self, *, slice_object: Slice,
                as_self: bool = True) -> Tuple[Status, Union[SliceManagerException, List[Sliver]]]:
        """
        Get slivers
        @param slice_object list of the slices
        @param as_self
        @return Tuple containing Status and Exception/Json containing Sliver(s)
        """
        if slice_object is None or not isinstance(slice_object, Slice):
            return Status.INVALID_ARGUMENTS, SliceManagerException("Invalid arguments - slice_object")

        try:
            if self.__should_renew():
                self.__load_tokens()

            return self.oc_proxy.slivers(token=self.get_id_token(), slice_id=slice_object.slice_id, as_self=as_self)
        except Exception as e:
            error_message = Utils.extract_error_message(exception=e)
            return Status.FAILURE, SliceManagerException(error_message)

    def resources(self, *, level: int = 1, force_refresh: bool = False, start: datetime = None, end: datetime = None,
                  includes: List[str] = None,
                  excludes: List[str] = None) -> Tuple[Status, Union[SliceManagerException, AdvertisedTopology]]:
        """
        Get resources
        @param level level
        @param force_refresh force_refresh
        @param start start time
        @param end end time
        @param includes list of sites to include
        @param excludes list of sites to exclude
        @return Tuple containing Status and Exception/Json containing Resources
        """
        try:
            if self.__should_renew():
                self.__load_tokens()
            return self.oc_proxy.resources(token=self.get_id_token(), level=level, force_refresh=force_refresh,
                                           start=start, end=end, includes=includes, excludes=excludes)
        except Exception as e:
            error_message = Utils.extract_error_message(exception=e)
            return Status.FAILURE, SliceManagerException(error_message)

    def renew(self, *, slice_object: Slice,
              new_lease_end_time: str) -> Tuple[Status, Union[SliceManagerException, None]]:
        """
        Renew a slice
        @param slice_object slice to be renewed
        @param new_lease_end_time new_lease_end_time
        @return Tuple containing Status and List of Reservation Id failed to extend
       """
        if slice_object is None or not isinstance(slice_object, Slice) or new_lease_end_time is None:
            return Status.INVALID_ARGUMENTS, SliceManagerException("Invalid arguments - "
                                                                   "slice_object or new_lease_end_time")
        try:
            if self.__should_renew():
                self.__load_tokens()

            return self.oc_proxy.renew(token=self.get_id_token(), slice_id=slice_object.slice_id,
                                       new_lease_end_time=new_lease_end_time)
        except Exception as e:
            error_message = Utils.extract_error_message(exception=e)
            return Status.FAILURE, SliceManagerException(error_message)

    def poa(self, *, sliver_id: str, operation: str, vcpu_cpu_map: List[Dict[str, str]] = None,
            node_set: List[str] = None,
            keys: List[Dict[str, str]] = None) ->Tuple[Status, Union[SliceManagerException, List[PoaData]]]:
        """
        Issue POA for a sliver
        @param sliver_id sliver Id for which to trigger POA
        @param operation operation
        @param vcpu_cpu_map list of mappings from virtual CPU to physical cpu
        @param node_set list of the numa nodes
        @param keys list of keys to add/remove
        @return Tuple containing Status and POA information
       """
        if sliver_id is None or operation is None:
            return Status.INVALID_ARGUMENTS, SliceManagerException("Invalid arguments - sliver_id or operation")

        try:
            if self.__should_renew():
                self.__load_tokens()

            return self.oc_proxy.poa(token=self.get_id_token(), sliver_id=sliver_id, operation=operation,
                                     vcpu_cpu_map=vcpu_cpu_map, node_set=node_set, keys=keys)

        except Exception as e:
            error_message = Utils.extract_error_message(exception=e)
            return Status.FAILURE, SliceManagerException(error_message)

    def get_poas(self, sliver_id: str = None, poa_id: str = None, limit: int = 20,
                 offset: int = 0, ) -> Tuple[Status, Union[SliceManagerException, List[PoaData]]]:
        """
        Get POAs
        @param sliver_id sliver Id for which to trigger POA
        @param limit maximum number of slices to return
        @param offset offset of the first slice to return
        @param poa_id POA id identifying the POA
        @return Tuple containing Status and POA information
        """
        try:
            if self.__should_renew():
                self.__load_tokens()
            return self.oc_proxy.get_poas(token=self.get_id_token(), limit=limit, offset=offset,
                                          sliver_id=sliver_id, poa_id=poa_id)
        except Exception as e:
            error_message = Utils.extract_error_message(exception=e)
            return Status.FAILURE, SliceManagerException(error_message)

    @staticmethod
    def __get_ssh_client() -> paramiko.SSHClient():
        client = paramiko.SSHClient()
        client.load_system_host_keys()
        client.set_missing_host_key_policy(paramiko.MissingHostKeyPolicy())
        client.set_missing_host_key_policy(paramiko.AutoAddPolicy())
        return client

    @staticmethod
    def execute(*, ssh_key_file: str, sliver: Node, username: str,
                command: str) -> Tuple[Status, SliceManagerException or Tuple]:
        """
        Execute a command on a sliver
        @param ssh_key_file: Location of SSH Private Key file to use to access the Sliver
        @param sliver: Node sliver
        @param username: Username to use to access the sliver
        @param command: Command to be executed on the sliver
        @return tuple as explained below:
        - Success: Status.OK and the stdout, and stderr of the executing command, as a 2-tuple
        - Failure: Status.Failure and exception
        Status indicates if the command could be executed(Status.OK) or not(Status.FAILURE).
        Success or failure of the command should be determined from the stdin, stdout and stderr
        """
        if sliver is None or not isinstance(sliver, Node) or ssh_key_file is None or\
                username is None or command is None:
            return Status.INVALID_ARGUMENTS, SliceManagerException("Invalid arguments - sliver or "
                                                                   "ssh_key_file or username or command")

        client = None
        try:
            key = paramiko.RSAKey.from_private_key_file(ssh_key_file)
            client = SliceManager.__get_ssh_client()
            client.connect(str(sliver.management_ip), username=username, pkey=key)
            stdin, stdout, stderr = client.exec_command(command=command)
            return Status.OK, (stdout.readlines(), stderr.readlines())
        except Exception as e:
            error_message = Utils.extract_error_message(exception=e)
            return Status.FAILURE, SliceManagerException(error_message)
        finally:
            if client is not None:
                client.close()

    def get_ssh_keys(self, uuid: str = None, email: str = None) -> list:
        """
        Return SSH Keys
        :return list of ssh keys
        """
        try:
            if self.__should_renew():
                self.__load_tokens()
            core_api_proxy = CoreApi(core_api_host=self.core_api_host, token=self.get_id_token())
            return core_api_proxy.get_ssh_keys(uuid=uuid, email=email)
        except Exception as e:
            error_message = Utils.extract_error_message(exception=e)
            raise SliceManagerException(error_message)

    def create_ssh_keys(self, key_type: str, description: str, comment: str = "ssh-key-via-api",
                        store_pubkey: bool = True) -> list:
        """
        Create SSH Keys for a user
        :param description: Key Description
        :param comment: Comment
        :param store_pubkey: Flag indicating if public key should be saved
        :param key_type: Key Type (sliver or bastion)

        :return list of ssh keys
        """
        try:
            if self.__should_renew():
                self.__load_tokens()
            core_api_proxy = CoreApi(core_api_host=self.core_api_host, token=self.get_id_token())
            return core_api_proxy.create_ssh_keys(key_type=key_type, comment=comment, store_pubkey=store_pubkey,
                                                  description=description)
        except Exception as e:
            error_message = Utils.extract_error_message(exception=e)
            raise SliceManagerException(error_message)

    def get_user_info(self, uuid: str = None, email: str = None) -> dict:
        """
        Return User's uuid by querying via Core API

        @return User's information
        """
        try:
            if self.__should_renew():
                self.__load_tokens()
            core_api_proxy = CoreApi(core_api_host=self.core_api_host, token=self.get_id_token())
            return core_api_proxy.get_user_info(uuid=uuid, email=email)
        except Exception as e:
            error_message = Utils.extract_error_message(exception=e)
            raise SliceManagerException(error_message)

    def get_project_info(self, project_name: str = "all", project_id: str = "all", uuid: str = None) -> list:
        """
        Get User's projects either identified by project name, project id or all
        @param project_id: Project Id
        @param project_name Project name
        @param uuid User Id

        @return list of projects
        """
        try:
            if self.__should_renew():
                self.__load_tokens()
            core_api_proxy = CoreApi(core_api_host=self.core_api_host, token=self.get_id_token())
            return core_api_proxy.get_user_projects(project_name=project_name, project_id=project_id, uuid=uuid)
        except Exception as e:
            error_message = Utils.extract_error_message(exception=e)
            raise SliceManagerException(error_message)

    def get_metrics_overview(self, excluded_projects: List[str] = None,
                             authenticated: bool = False) -> Tuple[Status, Union[list, Exception]]:
        """
        Get Metrics overview
        @param excluded_projects: excluded_projects
        @param authenticated: Specific user metrics
        @return list of metrics
        """
        try:
            token = None
            if authenticated and self.__should_renew():
                self.__load_tokens()
                token = self.get_id_token()
            return self.oc_proxy.get_metrics_overview(token=token, excluded_projects=excluded_projects)
        except Exception as e:
            error_message = Utils.extract_error_message(exception=e)
            return Status.FAILURE, SliceManagerException(error_message)


#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#
# Author: Erica Fu (ericafu@renci.org), Komal Thareja (kthare10@renci.org)
#
import os
import traceback
from typing import Any

import json
import click
from fabric_cf.orchestrator.orchestrator_proxy import SliceState
from fabric_cm.credmgr.credmgr_proxy import TokenType
from fabrictestbed.util.utils import Utils
from fim.slivers.network_node import NodeSliver

from .exceptions import TokenExpiredException
from ..slice_manager import CredmgrProxy
from ..slice_manager.slice_manager import SliceManager, Status
from ..util.constants import Constants


def __get_slice_manager(*, oc_host: str = None, cm_host: str = None, project_id: str = None, scope: str = "all",
                        token_location: str = None, project_name: str = None) -> SliceManager:
    """
    Get Environment Variables
    @param oc_host Orchestrator host
    @param cm_host Credmgr Host
    @param project_id Project Id
    @param project_name Project Name
    @param scope Scope
    @param token_location Absolute location of the tokens JSON file
    @raises ClickException in case of error
    """
    return SliceManager(oc_host=oc_host, cm_host=cm_host, project_id=project_id, scope=scope,
                        token_location=token_location, project_name=project_name)


def __unpack(data: Any) -> Any:
    """
    Recursivly unpacks JSON dictionaries or lists embedded in a list or dict.
    @param Any to unpack
    """
    if isinstance(data, str):
        if data and data[0] in ('{','['): ## starts with - json loads will catch errors
            data=__unpack(json.loads(data))
    elif isinstance(data, list):
        for i,v in enumerate(data):
            data[i]=__unpack(v)
    elif isinstance(data, dict):
        for k,v in data.items():
            data[k]=__unpack(v)
    return data


@click.group()
@click.option('-v', '--verbose', is_flag=True)
@click.pass_context
def cli(ctx, verbose):
    ctx.ensure_object(dict)
    ctx.obj['VERBOSE'] = verbose


@click.group()
@click.pass_context
def tokens(ctx):
    """ Token management
        (set $FABRIC_CREDMGR_HOST => CredentialManager, $FABRIC_project_id => Project Id)
    """


@tokens.command()
@click.option('--cmhost', help='Credmgr Host', default=None)
@click.option('--projectid', default=None, help='Project id')
@click.option('--projectname', default=None, help='Project name')
@click.option('--lifetime', default=4, help='Token lifetime in hours')
@click.option('--comment', default=None, help='Comment/note to associate with the token')
@click.option('--browser', default="chrome", type=click.Choice(['chrome', 'firefox', 'safari', 'edge'],
                                                               case_sensitive=False),
              help='Browser to use to launch the web app!')
@click.option('--scope', type=click.Choice(['cf', 'mf', 'all'], case_sensitive=False),
              default='all', help='scope')
@click.option('--tokenlocation', help='Location and name of the while in which to store tokens', default=None)
@click.pass_context
def create(ctx, cmhost: str, projectid: str, projectname: str, lifetime: int, comment: str,
           browser: str, scope: str, tokenlocation: str):
    """Create token
    """
    try:
        if cmhost is None and os.environ.get(Constants.FABRIC_CREDMGR_HOST) is None:
            raise click.ClickException(f"Credential Manager Host must be specified!")

        if projectname is None and os.environ.get(Constants.FABRIC_PROJECT_NAME) is None and projectid is None and \
                os.environ.get(Constants.FABRIC_PROJECT_ID) is None:
            raise click.ClickException(f"Either Project Name or Project Id must be specified!")

        cookie_name = os.getenv(Constants.FABRIC_COOKIE_NAME)
        if cmhost is None:
            cmhost = os.getenv(Constants.FABRIC_CREDMGR_HOST)

        if cookie_name is not None:
            cm_proxy = CredmgrProxy(credmgr_host=cmhost, cookie_name=cookie_name)
        else:
            cm_proxy = CredmgrProxy(credmgr_host=cmhost)

        if tokenlocation is None:
            tokenlocation = os.getenv(Constants.FABRIC_TOKEN_LOCATION)

        status, token_or_exception = cm_proxy.create(project_id=projectid, project_name=projectname,
                                                     life_time_in_hours=lifetime, comment=comment,
                                                     browser_name=browser, scope=scope, file_name=tokenlocation)

        from fabric_cm.credmgr.credmgr_proxy import Status as CmStatus
        if status == CmStatus.OK:
            if tokenlocation is None:
                click.echo(f"Fabric Token: {token_or_exception}")
            else:
                click.echo(f"Fabric Token saved at: {tokenlocation}")
        else:
            raise click.ClickException(f"{Utils.extract_error_message(exception=token_or_exception)}")
    except click.ClickException as e:
        raise e
    except Exception as e:
        raise click.ClickException(str(e))


@tokens.command()
@click.option('--cmhost', help='Credmgr Host', default=None)
@click.option('--tokenlocation', help='Location of the file which contains the valid tokens', default=None)
@click.option('--projectid', default=None, help='Project Id')
@click.option('--projectname', default=None, help='Project name')
@click.option('--scope', type=click.Choice(['cf', 'mf', 'all'], case_sensitive=False),
              default='all', help='Scope')
@click.pass_context
def refresh(ctx, cmhost: str, tokenlocation: str, projectid: str, projectname: str, scope: str):
    """Refresh token
    """
    try:
        if cmhost is None and os.environ.get(Constants.FABRIC_CREDMGR_HOST) is None:
            raise click.ClickException(f"Credential Manager Host must be specified!")

        if tokenlocation is None and os.environ.get(Constants.FABRIC_TOKEN_LOCATION) is None:
            raise click.ClickException(f"Token location must be specified !")

        if projectname is None and os.environ.get(Constants.FABRIC_PROJECT_NAME) is None and projectid is None and \
                os.environ.get(Constants.FABRIC_PROJECT_ID) is None:
            raise click.ClickException(f"Either Project Name or Project Id must be specified!")

        if tokenlocation is None:
            tokenlocation = os.getenv(Constants.FABRIC_TOKEN_LOCATION)

        slice_manager = __get_slice_manager(cm_host=cmhost, project_id=projectid, scope=scope,
                                            token_location=tokenlocation, project_name=projectname)

        click.echo(f"ID Token: {slice_manager.get_id_token()}")
        click.echo(f"Refresh Token: {slice_manager.get_refresh_token()}")
        click.echo(f"Refreshed token saved at: {tokenlocation}")
    except click.ClickException as e:
        raise e
    except Exception as e:
        raise click.ClickException(str(e))


@tokens.command()
@click.option('--cmhost', help='Credmgr Host', default=None)
@click.option('--tokenlocation', help='Location of the file which contains the valid tokens', default=None)
@click.option('--refreshtoken', help='Refresh token to be revoked; '
                                     'If specified refresh token from tokenlocation is ignored!', default=None)
@click.option('--identitytoken', help='Identity token to be revoked or to authenticate the user when revoking a '
                                      'refresh token; If specified identity token from tokenlocation is ignored',
              default=None)
@click.option('--tokenhash', help='SHA256 hash for the token being revoked; If not specified, '
                                  'it is assumed that the identity token being passed is being revoked', default=None)
@click.pass_context
def revoke(ctx, cmhost: str, tokenlocation: str, refreshtoken: str, identitytoken: str, tokenhash: str):
    """ Revoke token
    """
    try:
        fail = False
        if cmhost is None and os.environ.get(Constants.FABRIC_CREDMGR_HOST) is None:
            raise click.ClickException(f"Credential Manager Host must be specified!")

        if tokenlocation is None and os.environ.get(Constants.FABRIC_TOKEN_LOCATION) is None and \
                refreshtoken is None and identitytoken is None:
            raise click.ClickException(f"Either Token location must be specified or pass refresh "
                                       f"token and identity token!")

        # Token in the file located at tokenlocation is being revoked
        if tokenlocation is not None or os.environ.get(Constants.FABRIC_TOKEN_LOCATION) is not None:
            if tokenlocation is None:
                tokenlocation = os.getenv(Constants.FABRIC_TOKEN_LOCATION)

            if os.path.exists(tokenlocation):
                with open(tokenlocation, 'r') as stream:
                    tokens = json.loads(stream.read())
                    refreshtoken = tokens.get(CredmgrProxy.REFRESH_TOKEN)
                    identitytoken = tokens.get(CredmgrProxy.ID_TOKEN)
                    tokenhash = tokens.get("token_hash")
            else:
                raise click.ClickException(f"Token file '{tokenlocation}' does not exist!")

        if cmhost is None:
            cmhost = os.environ.get(Constants.FABRIC_CREDMGR_HOST)

        cm_proxy = CredmgrProxy(credmgr_host=cmhost)
        if refreshtoken is None:
            status, error_str = cm_proxy.revoke(identity_token=identitytoken, token_hash=tokenhash,
                                                token_type=TokenType.Identity)
        else:

            status, error_str = cm_proxy.revoke(refresh_token=refreshtoken, identity_token=identitytoken,
                                                token_type=TokenType.Refresh)
        from fabric_cm.credmgr.credmgr_proxy import Status as CmStatus
        if status == CmStatus.OK:
            click.echo("Token revoked successfully")
        else:
            raise click.ClickException(f"{Utils.extract_error_message(exception=error_str)}")
    except click.ClickException as e:
        raise e
    except Exception as e:
        traceback.print_exc()
        raise click.ClickException(str(e))


@tokens.command()
@click.option('--cmhost', help='Credmgr Host', default=None)
@click.option('--tokenlocation', help='location for the tokens', default=None)
@click.pass_context
def clear_cache(ctx, cmhost, tokenlocation):
    """ Clear cached token
    """
    try:
        if cmhost is None and os.environ.get(Constants.FABRIC_CREDMGR_HOST) is None:
            raise click.ClickException(f"Credential Manager Host must be specified!")

        if tokenlocation is None:
            raise click.ClickException(f"Token location must be specified !")

        slice_manager = __get_slice_manager(cm_host=cmhost, token_location=tokenlocation)

        status, error_str = slice_manager.clear_token_cache(file_name=tokenlocation)
        if status == Status.OK:
            click.echo("Token cache cleared successfully")
        else:
            raise click.ClickException(f"{Utils.extract_error_message(exception=error_str)}")

    except click.ClickException as e:
        raise e
    except Exception as e:
        raise click.ClickException(str(e))


@click.group()
@click.pass_context
def slices(ctx):
    """ Slice management
        (set $FABRIC_ORCHESTRATOR_HOST => Orchestrator, $FABRIC_CREDMGR_HOST => CredentialManager,
        $FABRIC_TOKEN_LOCATION => Location of the token file, $FABRIC_PROJECT_ID => Project Id)
    """


@slices.command()
@click.option('--cmhost', help='Credmgr Host', default=None)
@click.option('--ochost', help='Orchestrator Host', default=None)
@click.option('--tokenlocation', help='location for the tokens', default=None)
@click.option('--projectid', default=None, help='project name')
@click.option('--scope', type=click.Choice(['cf', 'mf', 'all'], case_sensitive=False),
              default='all', help='scope')
@click.option('--sliceid', default=None, help='Slice Id')
@click.option('--state', default=None, help='Slice State')
@click.pass_context
def query(ctx, cmhost: str, ochost: str, tokenlocation: str, projectid: str, scope: str, sliceid: str, state: str):
    """ Query slice_editor slice(s)
    """
    try:
        slice_manager = __get_slice_manager(cm_host=cmhost, oc_host=ochost, project_id=projectid, scope=scope,
                                            token_location=tokenlocation)
        status = None
        response = None
        includes = []
        if state is not None:
            slice_state = SliceState.state_from_str(state)
            if slice_state is not None:
                includes.append(slice_state)

        status, response = slice_manager.slices(includes=includes, slice_id=sliceid)

        if status == Status.OK and not isinstance(response, Exception):
            click.echo(json.dumps(list(map(lambda i: i.to_dict(), response)),indent=2))
        else:
            click.echo(Utils.extract_error_message(exception=response))

    except TokenExpiredException:
        raise click.ClickException("Unauthorized: Valid token required")
    except Exception as e:
        raise click.ClickException(str(e))


@slices.command()
@click.option('--cmhost', help='Credmgr Host', default=None)
@click.option('--ochost', help='Orchestrator Host', default=None)
@click.option('--tokenlocation', help='location for the tokens', default=None)
@click.option('--projectid', default=None, help='project name')
@click.option('--scope', type=click.Choice(['cf', 'mf', 'all'], case_sensitive=False),
              default='all', help='scope')
@click.option('--slicename', help='Slice Name', required=True)
@click.option('--slicegraph', help='Slice Graph', required=True)
@click.option('--sshkey', help='SSH Key', required=True)
@click.option('--leaseend', help='Lease End', default=None)
@click.pass_context
def create(ctx, cmhost: str, ochost: str, tokenlocation: str, projectid: str, scope: str, slicename: str,
           slicegraph: str, sshkey: str, leaseend: str):
    """ Create slice_editor slice
    """
    try:
        slice_manager = __get_slice_manager(cm_host=cmhost, oc_host=ochost, project_id=projectid, scope=scope,
                                            token_location=tokenlocation)
        status, response = slice_manager.create(slice_name=slicename, slice_graph=slicegraph, ssh_key=sshkey,
                                                lease_end_time=leaseend)

        if status == Status.OK:
            click.echo(response)
        else:
            click.echo(Utils.extract_error_message(exception=response))

    except TokenExpiredException:
        raise click.ClickException("Unauthorized: Valid token required")
    except Exception as e:
        raise click.ClickException(str(e))


@slices.command()
@click.option('--cmhost', help='Credmgr Host', default=None)
@click.option('--ochost', help='Orchestrator Host', default=None)
@click.option('--tokenlocation', help='location for the tokens', default=None)
@click.option('--projectid', default=None, help='project name')
@click.option('--scope', type=click.Choice(['cf', 'mf', 'all'], case_sensitive=False),
              default='all', help='scope')
@click.option('--sliceid', help='Slice Id', required=True)
@click.option('--slicegraph', help='Slice Graph', required=True)
@click.pass_context
def modify(ctx, cmhost: str, ochost: str, tokenlocation: str, projectid: str, scope: str, sliceid: str,
           slicegraph: str):
    """ Modify an existing slice
    """
    try:
        slice_manager = __get_slice_manager(cm_host=cmhost, oc_host=ochost, project_id=projectid, scope=scope,
                                            token_location=tokenlocation)
        status, response = slice_manager.modify(slice_id=sliceid, slice_graph=slicegraph)

        if status == Status.OK:
            click.echo(response)
        else:
            click.echo(Utils.extract_error_message(exception=response))

    except TokenExpiredException:
        raise click.ClickException("Unauthorized: Valid token required")
    except Exception as e:
        raise click.ClickException(str(e))


@slices.command()
@click.option('--cmhost', help='Credmgr Host', default=None)
@click.option('--ochost', help='Orchestrator Host', default=None)
@click.option('--tokenlocation', help='location for the tokens', default=None)
@click.option('--projectid', default=None, help='project name')
@click.option('--scope', type=click.Choice(['cf', 'mf', 'all'], case_sensitive=False),
              default='all', help='scope')
@click.option('--sliceid', help='Slice Id', required=True)
@click.pass_context
def modifyaccept(ctx, cmhost: str, ochost: str, tokenlocation: str, projectid: str, scope: str, sliceid: str):
    """ Accept the modified slice
    """
    try:
        slice_manager = __get_slice_manager(cm_host=cmhost, oc_host=ochost, project_id=projectid, scope=scope,
                                            token_location=tokenlocation)
        status, response = slice_manager.modify_accept(slice_id=sliceid)

        if status == Status.OK:
            click.echo(response)
        else:
            click.echo(Utils.extract_error_message(exception=response))

    except TokenExpiredException:
        raise click.ClickException("Unauthorized: Valid token required")
    except Exception as e:
        raise click.ClickException(str(e))


@slices.command()
@click.option('--cmhost', help='Credmgr Host', default=None)
@click.option('--ochost', help='Orchestrator Host', default=None)
@click.option('--tokenlocation', help='location for the tokens', default=None)
@click.option('--projectid', default=None, help='project name')
@click.option('--scope', type=click.Choice(['cf', 'mf', 'all'], case_sensitive=False),
              default='all', help='scope')
@click.option('--sliceid', help='Slice Id', required=False)
@click.pass_context
def delete(ctx, cmhost: str, ochost: str, tokenlocation: str, projectid: str, scope: str, sliceid: str):
    """ Delete slice_editor slice
    """
    try:
        slice_manager = __get_slice_manager(cm_host=cmhost, oc_host=ochost, project_id=projectid, scope=scope,
                                            token_location=tokenlocation)
        slice_object = None
        if sliceid is not None:
            status, response = slice_manager.slices(slice_id=sliceid)
            if status != Status.OK or isinstance(response, Exception):
                click.echo(f'Delete Slice failed: {status.interpret(exception=response)}')
                return
            slice_object = response[0]

        status, response = slice_manager.delete(slice_object=slice_object)

        if status == Status.OK:
            click.echo(response)
        else:
            click.echo(Utils.extract_error_message(exception=response))

    except TokenExpiredException:
        raise click.ClickException("Unauthorized: Valid token required")
    except Exception as e:
        raise click.ClickException(str(e))


@click.group()
@click.pass_context
def slivers(ctx):
    """ Sliver management
        (set $FABRIC_ORCHESTRATOR_HOST => Orchestrator, $FABRIC_CREDMGR_HOST => CredentialManager,
        $FABRIC_TOKEN_LOCATION => Location of the token file, $FABRIC_PROJECT_ID => Project Id)
    """


@slivers.command()
@click.option('--cmhost', help='Credmgr Host', default=None)
@click.option('--ochost', help='Orchestrator Host', default=None)
@click.option('--tokenlocation', help='location for the tokens', default=None)
@click.option('--projectid', default=None, help='project name')
@click.option('--scope', type=click.Choice(['cf', 'mf', 'all'], case_sensitive=False),
              default='all', help='scope')
@click.option('--sliceid', help='Slice Id')
@click.option('--sliverid', default=None, help='Sliver Id')
@click.pass_context
def query(ctx, cmhost: str, ochost: str, tokenlocation: str, projectid: str, scope: str, sliceid: str, sliverid: str):
    """ Query slice_editor slice sliver(s)
    """
    try:
        slice_manager = __get_slice_manager(cm_host=cmhost, oc_host=ochost, project_id=projectid, scope=scope,
                                            token_location=tokenlocation)

        status, response = slice_manager.slices(slice_id=sliceid)
        if status != Status.OK:
            click.echo(f'Query Sliver(s) failed: {status.interpret(exception=response)}')
            return

        slice_object = response[0]
        status, response = slice_manager.slivers(slice_object=slice_object)

        if status == Status.OK and not isinstance(response, Exception):
            click.echo(json.dumps(list(map(lambda i: __unpack(i.to_dict()), response)),indent=2))
        else:
            click.echo(Utils.extract_error_message(exception=response))

    except TokenExpiredException:
        raise click.ClickException("Unauthorized: Valid token required")
    except Exception as e:
        raise click.ClickException(str(e))


@slivers.command()
@click.option('--sshkeyfile', help='Location of SSH Private Key file to use to access the Sliver')
@click.option('--sliceaddress', help='IP address or the connection string to use to access the sliver')
@click.option('--username', default=None, help='Username to use to access the sliver')
@click.option('--command', default=None, help='Command to be executed on the sliver')
@click.pass_context
def execute(ctx, sshkeyfile: str, sliceaddress: str, username: str, command: str):
    """ Query slice_editor slice sliver(s)
    """
    try:
        slice_manager = __get_slice_manager()

        sliver = NodeSliver()
        sliver.management_ip = sliceaddress
        status, response = slice_manager.execute(ssh_key_file=sshkeyfile, sliver=sliver,
                                                 username=username, command=command)

        if status == Status.OK:
            output, error = response
            click.echo(f"Output: {output}")
            click.echo(f"Error: {error}")
        else:
            click.echo(Utils.extract_error_message(exception=response))
    except Exception as e:
        raise click.ClickException(str(e))


@click.group()
@click.pass_context
def resources(ctx):
    """ Resource management
        (set $FABRIC_ORCHESTRATOR_HOST => Orchestrator, $FABRIC_CREDMGR_HOST => CredentialManager, $FABRIC_PROJECT_ID => Project Id)
    """


@resources.command()
@click.option('--cmhost', help='Credmgr Host', default=None)
@click.option('--ochost', help='Orchestrator Host', default=None)
@click.option('--tokenlocation', help='location for the tokens', default=None)
@click.option('--projectid', default=None, help='project name')
@click.option('--scope', type=click.Choice(['cf', 'mf', 'all'], case_sensitive=False),
              default='all', help='scope')
@click.option('--force', default=False, help='Force current snapshot')
@click.pass_context
def query(ctx, cmhost: str, ochost: str, tokenlocation: str, projectid: str, scope: str, force: bool):
    """ Query resources
    """
    try:
        slice_manager = __get_slice_manager(cm_host=cmhost, oc_host=ochost, project_id=projectid, scope=scope,
                                            token_location=tokenlocation)

        status, response = slice_manager.resources(force_refresh=force)

        if status == Status.OK:
            click.echo(response)
        else:
            click.echo(Utils.extract_error_message(exception=response))

    except TokenExpiredException:
        raise click.ClickException("Unauthorized: Valid token required")
    except Exception as e:
        raise click.ClickException(str(e))


cli.add_command(tokens)
cli.add_command(slices)
cli.add_command(slivers)
cli.add_command(resources)

